/*
 * Copyright (c) 2001,2002 Florian Schulze.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of the contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * test.c - This file is part of lwIP test
 *
 */

/* C runtime includes */
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <conio.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* lwIP core includes */
#include "lwip/opt.h"

#include "lwip/sys.h"
#include "lwip/timeouts.h"
#include "lwip/debug.h"
#include "lwip/stats.h"
#include "lwip/init.h"
#include "lwip/tcpip.h"
#include "lwip/netif.h"
#include "lwip/api.h"

#include "lwip/tcp.h"
#include "lwip/udp.h"
#include "lwip/dns.h"
#include "lwip/dhcp.h"
#include "lwip/autoip.h"
#include "lwip/priv/tcp_priv.h"

/* lwIP netif includes */
#include "lwip/etharp.h"
#include "netif/ethernet.h"
#include "netif/ppp/ppp.h"

/* applications includes */
#include "lwip/apps/lwiperf.h"
#include "lwip/apps/netbiosns.h"
#include "lwip/apps/sntp.h"
#include "lwip/apps/snmp.h"
#include "lwip/apps/snmp_mib2.h"
#include "lwip/apps/httpd.h"
#include "lwip/apps/mdns.h"
#include "lwip/apps/tftp_server.h"
#include "lwip/apps/tftp_client.h"
#include "apps/httpserver/httpserver-netconn.h"
#include "apps/netio/netio.h"
#include "apps/ping/ping.h"
#include "apps/rtp/rtp.h"
#include "apps/chargen/chargen.h"
#include "apps/shell/shell.h"
#include "apps/tcpecho/tcpecho.h"
#include "apps/udpecho/udpecho.h"
#include "apps/tcpecho_raw/tcpecho_raw.h"
#include "apps/socket_examples/socket_examples.h"
#include "examples/snmp/snmp_v3/snmpv3_dummy.h"

#include "examples/lwiperf/lwiperf_example.h"
#include "examples/mdns/mdns_example.h"
#include "examples/snmp/snmp_example.h"
#include "examples/tftp/tftp_example.h"
#include "examples/sntp/sntp_example.h"
#include "examples/mqtt/mqtt_example.h"

#include "examples/httpd/cgi_example/cgi_example.h"
#include "examples/httpd/fs_example/fs_example.h"
#include "examples/httpd/ssi_example/ssi_example.h"

#if NO_SYS
  /* ... then we need information about the timer intervals: */
  #include "lwip/ip4_frag.h"
  #include "lwip/igmp.h"
#endif

#include "netif/ppp/ppp_opts.h"
#include "netif/bridgeif.h"

#if USE_PPP || PPP_SUPPORT
  /* PPP includes */
  #include "lwip/sio.h"
  #include "netif/ppp/pppapi.h"
  #include "netif/ppp/pppos.h"
  #include "netif/ppp/pppoe.h"

  #if !NO_SYS && !LWIP_PPP_API
    #error With NO_SYS==0, LWIP_PPP_API==1 is required.
  #endif
#endif

#if LWIP_SNMP
  /* SNMP trap destination cmd option */
  static unsigned char trap_flag = 1;
  static ip_addr_t     trap_addr;

  /* 'non-volatile' SNMP settings
    @todo: make these truly non-volatile */
  u8_t  syscontact_str[255];
  u16_t syscontact_len = 0;
  u8_t  syslocation_str[255];
  u16_t syslocation_len = 0;

  /* enable == 1, disable == 2 */
  u8_t snmpauthentraps_set = 1;
#endif

/** Define this to 1 to enable a PCAP interface as default interface. */
#ifndef USE_PCAPIF
#define USE_PCAPIF 1
#endif

/** Define this to 1 to enable a PPP interface. */
#ifndef USE_PPP
#define USE_PPP 0
#endif

/** Define this to 1 or 2 to support 1 or 2 SLIP interfaces. */
#ifndef USE_SLIPIF
#define USE_SLIPIF 0
#endif

/** Use an ethernet adapter? Default to enabled if PCAPIF or PPPoE are used. */
#ifndef USE_ETHERNET
#define USE_ETHERNET  (USE_PCAPIF || PPPOE_SUPPORT)
#endif

/** Use an ethernet adapter for TCP/IP? By default only if PCAPIF is used. */
#ifndef USE_ETHERNET_TCPIP
#define USE_ETHERNET_TCPIP  (USE_PCAPIF)
#endif

#if USE_ETHERNET
#include "pcapif.h"
#endif

#if USE_SLIPIF
#include <netif/slipif.h>
#endif

#ifndef USE_DHCP
#define USE_DHCP    LWIP_DHCP
#endif

#ifndef USE_AUTOIP
#define USE_AUTOIP  LWIP_AUTOIP
#endif

/* globales variables for netifs
 */
#if USE_ETHERNET
  /* THE ethernet interface */
  struct netif netif;

  /* The bridge interface */
  struct netif bridge;

  #if LWIP_DHCP
  /* dhcp struct for the ethernet netif */
  struct dhcp netif_dhcp;
  #endif

  #if LWIP_AUTOIP
  /* autoip struct for the ethernet netif */
  struct autoip netif_autoip;
  #endif
#endif

#if USE_PPP
  /* THE PPP PCB */
  ppp_pcb *ppp;
  /* THE PPP interface */
  struct netif ppp_netif;
  /* THE PPP descriptor */
  u8_t sio_idx = 0;
  sio_fd_t ppp_sio;
#endif

#if USE_SLIPIF
  struct netif slipif1;
  #if USE_SLIPIF > 1
    struct netif slipif2;
  #endif
#endif

#if LWIP_UDP

static void *tftp_open(const char* fname, const char* mode, u8_t is_write)
{
  void *rc;

  LWIP_UNUSED_ARG(mode);
  if (is_write)
       rc = (void*)fopen(fname, "wb");
  else rc = (void*)fopen(fname, "rb");

  LWIP_TRACE (1, "%s(): fname: '%s', mode: '%s', is_write: %d, rc: %p\n", __FUNCTION__, fname, mode, is_write, rc);
  return (rc);
}

static void tftp_close(void* handle)
{
  LWIP_TRACE (1, "%s(): handle: %p\n", __FUNCTION__, handle);
  fclose((FILE*)handle);
}

static int tftp_read(void* handle, void* buf, int bytes)
{
  int ret = (int)fread (buf, 1, bytes, (FILE*)handle);

  LWIP_TRACE (1, "%s(): bytes: %d, ret: %d\n", __FUNCTION__, bytes, ret);

  if (ret <= 0)
     return (-1);
  return (ret);
}

static int tftp_write(void* handle, struct pbuf* p)
{
  LWIP_TRACE (1, "%s(): handle: %p\n", __FUNCTION__, handle);

  while (p != NULL)
  {
    if (fwrite(p->payload, 1, p->len, (FILE*)handle) != (size_t)p->len)
       return -1;
    p = p->next;
  }
  return 0;
}

static const struct tftp_context tftp = {
  tftp_open,
  tftp_close,
  tftp_read,
  tftp_write
};
#endif

#if USE_PPP
static void
pppLinkStatusCallback(ppp_pcb *pcb, int errCode, void *ctx)
{
  struct netif *pppif = ppp_netif(pcb);
  LWIP_UNUSED_ARG(ctx);

  switch(errCode) {
    case PPPERR_NONE:              /* No error. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_NONE\n");
#if LWIP_IPV4
      LWIP_TRACE (1, "   our_ipaddr  = %s\n", ip4addr_ntoa(netif_ip4_addr(pppif)));
      LWIP_TRACE (1, "   his_ipaddr  = %s\n", ip4addr_ntoa(netif_ip4_gw(pppif)));
      LWIP_TRACE (1, "   netmask     = %s\n", ip4addr_ntoa(netif_ip4_netmask(pppif)));
#endif

#if LWIP_DNS
      LWIP_TRACE (1, "   dns1        = %s\n", ipaddr_ntoa(dns_getserver(0)));
      LWIP_TRACE (1, "   dns2        = %s\n", ipaddr_ntoa(dns_getserver(1)));
#endif

#if PPP_IPV6_SUPPORT
      LWIP_TRACE (1, "   our6_ipaddr = %s\n", ip6addr_ntoa(netif_ip6_addr(pppif, 0)));
#endif
      break;

    case PPPERR_PARAM: {          /* Invalid parameter. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_PARAM\n");
      break;
    }
    case PPPERR_OPEN: {           /* Unable to open PPP session. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_OPEN\n");
      break;
    }
    case PPPERR_DEVICE: {         /* Invalid I/O device for PPP. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_DEVICE\n");
      break;
    }
    case PPPERR_ALLOC: {          /* Unable to allocate resources. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_ALLOC\n");
      break;
    }
    case PPPERR_USER: {           /* User interrupt. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_USER\n");
      break;
    }
    case PPPERR_CONNECT: {        /* Connection lost. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_CONNECT\n");
      break;
    }
    case PPPERR_AUTHFAIL: {       /* Failed authentication challenge. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_AUTHFAIL\n");
      break;
    }
    case PPPERR_PROTOCOL: {       /* Failed to meet protocol. */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_PROTOCOL\n");
      break;
    }
    case PPPERR_PEERDEAD: {       /* Connection timeout */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_PEERDEAD\n");
      break;
    }
    case PPPERR_IDLETIMEOUT: {    /* Idle Timeout */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_IDLETIMEOUT\n");
      break;
    }
    case PPPERR_CONNECTTIME: {    /* Max connect time reached */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_CONNECTTIME\n");
      break;
    }
    case PPPERR_LOOPBACK: {       /* Loopback detected */
      LWIP_TRACE (1, "pppLinkStatusCallback: PPPERR_LOOPBACK\n");
      break;
    }
    default: {
      LWIP_TRACE (1, "pppLinkStatusCallback: unknown errCode %d\n", errCode);
      break;
    }
  }
}

#if PPPOS_SUPPORT
static u32_t
ppp_output_cb(ppp_pcb *pcb, u8_t *data, u32_t len, void *ctx)
{
  LWIP_UNUSED_ARG(pcb);
  LWIP_UNUSED_ARG(ctx);
  return sio_write(ppp_sio, data, len);
}
#endif
#endif /* USE_PPP */

#if LWIP_NETIF_STATUS_CALLBACK
static void
status_callback(struct netif *state_netif)
{
  if (netif_is_up(state_netif)) {
#if LWIP_IPV4
    LWIP_TRACE (1, "status_callback==UP, local interface IP is %s\n", ip4addr_ntoa(netif_ip4_addr(state_netif)));
#else
    LWIP_TRACE (1, "status_callback==UP\n");
#endif

#if LWIP_MDNS_RESPONDER
    mdns_resp_netif_settings_changed(state_netif);
#endif
  }
  else {
    LWIP_TRACE (1, "status_callback==DOWN\n");
  }
}
#endif /* LWIP_NETIF_STATUS_CALLBACK */

#if LWIP_NETIF_LINK_CALLBACK
static void
link_callback(struct netif *state_netif)
{
  LWIP_TRACE (1, "link_callback == %s\n", netif_is_link_up(state_netif) ? "UP" : "DOWN");
}
#endif

/* This function initializes all network interfaces
 */
static void msvc_netif_init(void)
{
#if LWIP_IPV4 && USE_ETHERNET
  ip4_addr_t ipaddr, netmask, gw;
#endif

#if USE_SLIPIF
  u8_t num_slip1 = 0;

  #if LWIP_IPV4
    ip4_addr_t ipaddr_slip1, netmask_slip1, gw_slip1;
  #endif

  #if USE_SLIPIF > 1
    u8_t num_slip2 = 1;
    #if LWIP_IPV4
      ip4_addr_t ipaddr_slip2, netmask_slip2, gw_slip2;
    #endif
  #endif
#endif

#if USE_DHCP || USE_AUTOIP
  err_t err;
#endif

#if USE_PPP
  const char *username = NULL, *password = NULL;

  #ifdef PPP_USERNAME
    username = PPP_USERNAME;
  #endif

  #ifdef PPP_PASSWORD
    password = PPP_PASSWORD;
  #endif

  LWIP_TRACE (1, "ppp_connect: COM%d\n", (int)sio_idx);

  #if PPPOS_SUPPORT
    ppp_sio = sio_open(sio_idx);
    if (ppp_sio == NULL)
       LWIP_TRACE (1, "sio_open error\n");
    else
    {
      ppp = pppos_create(&ppp_netif, ppp_output_cb, pppLinkStatusCallback, NULL);
      if (ppp == NULL)
           LWIP_TRACE (1, "pppos_create error\n");
      else ppp_connect(ppp, 0);
    }
  #endif
#endif

#if USE_ETHERNET
  #if LWIP_IPV4
      #define NETIF_ADDRS &ipaddr, &netmask, &gw,
      ip4_addr_set_zero(&gw);
      ip4_addr_set_zero(&ipaddr);
      ip4_addr_set_zero(&netmask);
    #if USE_ETHERNET_TCPIP
      #if USE_DHCP
        LWIP_TRACE (1, "Starting lwIP, local interface IP is dhcp-enabled\n");
      #elif USE_AUTOIP
        LWIP_TRACE (1, "Starting lwIP, local interface IP is autoip-enabled\n");
      #else
        LWIP_PORT_INIT_GW(&gw);
        LWIP_PORT_INIT_IPADDR(&ipaddr);
        LWIP_PORT_INIT_NETMASK(&netmask);
        LWIP_TRACE (1, "Starting lwIP, local interface IP is %s\n", ip4addr_ntoa(&ipaddr));
      #endif
    #endif
  #else
    #define NETIF_ADDRS
    LWIP_TRACE (1, "Starting lwIP, IPv4 disable\n");
  #endif

  #if NO_SYS
    netif_set_default(netif_add(&netif, NETIF_ADDRS NULL, pcapif_init, netif_input));
  #else
    netif_set_default(netif_add(&netif, NETIF_ADDRS NULL, pcapif_init, tcpip_input));
  #endif

  #if (LWIP_NUM_NETIF_CLIENT_DATA > 1) || defined(USE_BRIDGE)
    bridgeif_init (&netif);
    bridgeif_add_port (&bridge, &netif);
  #endif

  #if LWIP_IPV6
    netif_create_ip6_linklocal_address(&netif, 1);
    LWIP_TRACE (1, "ip6 linklocal address: %s\n", ip6addr_ntoa (netif_ip6_addr(&netif, 0)));
  #endif

  #if LWIP_NETIF_STATUS_CALLBACK
    netif_set_status_callback(&netif, status_callback);
  #endif

  #if LWIP_NETIF_LINK_CALLBACK
    netif_set_link_callback(&netif, link_callback);
  #endif

  #if USE_ETHERNET_TCPIP
    #if LWIP_AUTOIP
      autoip_set_struct(&netif, &netif_autoip);
    #endif

    #if LWIP_DHCP
      dhcp_set_struct(&netif, &netif_dhcp);
    #endif

    netif_set_up(&netif);

    if (lwip_cfg.ip4.fixed_addr)
    {
    #if LWIP_IPV4
      netif_set_ipaddr (&netif, (const ip4_addr_t*)&lwip_cfg.ip4.fixed_addr);
      netif_set_netmask (&netif, (const ip4_addr_t*)&lwip_cfg.ip4.fixed_netmask);
      netif_set_gw (&netif, (const ip4_addr_t*)&lwip_cfg.ip4.fixed_gateway);
    #endif
    }
    else
    {
    #if USE_DHCP
      err = dhcp_start(&netif);
      LWIP_ASSERT("dhcp_start failed", err == ERR_OK);
    #elif USE_AUTOIP
      err = autoip_start(&netif);
      LWIP_ASSERT("autoip_start failed", err == ERR_OK);
    #endif
    }
  #else

    /* Use ethernet for PPPoE only */
    netif.flags &= ~(NETIF_FLAG_ETHARP | NETIF_FLAG_IGMP); /* no ARP */
    netif.flags |= NETIF_FLAG_ETHERNET; /* but pure ethernet */
  #endif

  #if USE_PPP && PPPOE_SUPPORT
    /* start PPPoE after ethernet netif is added! */
    ppp = pppoe_create(&ppp_netif, &netif, NULL, NULL, pppLinkStatusCallback, NULL);
    if (ppp == NULL)
      LWIP_TRACE (1, "pppos_create error\n");
    else {
      #if PPP_AUTH_SUPPORT
        ppp_set_auth(ppp, PPPAUTHTYPE_ANY, username, password);
      #endif
      ppp_connect(ppp, 0);
    }
  #endif
#endif /* USE_ETHERNET */

#if USE_SLIPIF
  #if LWIP_IPV4
    #define SLIP1_ADDRS &ipaddr_slip1, &netmask_slip1, &gw_slip1,
    LWIP_PORT_INIT_SLIP1_IPADDR(&ipaddr_slip1);
    LWIP_PORT_INIT_SLIP1_GW(&gw_slip1);
    LWIP_PORT_INIT_SLIP1_NETMASK(&netmask_slip1);
    LWIP_TRACE (1, "Starting lwIP slipif, local interface IP is %s\n", ip4addr_ntoa(&ipaddr_slip1));
  #else
    #define SLIP1_ADDRS
    LWIP_TRACE (1, "Starting lwIP slipif\n");
  #endif

  #if defined(SIO_USE_COMPORT) && SIO_USE_COMPORT
    num_slip1++; /* COM ports cannot be 0-based */
  #endif

  netif_add(&slipif1, SLIP1_ADDRS &num_slip1, slipif_init, ip_input);

  #if !USE_ETHERNET
    netif_set_default(&slipif1);
  #endif

  #if LWIP_IPV6
    netif_create_ip6_linklocal_address(&slipif1, 1);
  #endif

  #if LWIP_NETIF_STATUS_CALLBACK
    LWIP_TRACE (1, "SLIP ip6 linklocal address: %s\n",
                 ip4addr_ntoa(netif_ip6_addr(&slipif1, 0)));
    netif_set_status_callback(&slipif1, status_callback);
  #endif

  #if LWIP_NETIF_LINK_CALLBACK
    netif_set_link_callback(&slipif1, link_callback);
  #endif

  netif_set_up(&slipif1);

  #if USE_SLIPIF > 1
    #if LWIP_IPV4
      #define SLIP2_ADDRS &ipaddr_slip2, &netmask_slip2, &gw_slip2,

      LWIP_PORT_INIT_SLIP2_IPADDR(&ipaddr_slip2);
      LWIP_PORT_INIT_SLIP2_GW(&gw_slip2);
      LWIP_PORT_INIT_SLIP2_NETMASK(&netmask_slip2);
      LWIP_TRACE (1, "Starting lwIP SLIP if #2, local interface IP is %s\n", ip4addr_ntoa(&ipaddr_slip2));
    #else
      #define SLIP2_ADDRS
      LWIP_TRACE (1, "Starting lwIP SLIP if #2\n");
    #endif

    #if defined(SIO_USE_COMPORT) && SIO_USE_COMPORT
      num_slip2++; /* COM ports cannot be 0-based */
    #endif

    netif_add(&slipif2, SLIP2_ADDRS &num_slip2, slipif_init, ip_input);

    #if LWIP_IPV6
      netif_create_ip6_linklocal_address(&slipif1, 1);
      LWIP_TRACE (1, "SLIP2 ip6 linklocal address: %s\n", ip6addr_ntoa(netif_ip6_addr(&slipif2, 0)));
    #endif

    #if LWIP_NETIF_STATUS_CALLBACK
      netif_set_status_callback(&slipif2, status_callback);
    #endif

    #if LWIP_NETIF_LINK_CALLBACK
      netif_set_link_callback(&slipif2, link_callback);
    #endif

    netif_set_up(&slipif2);
  #endif
#endif /* USE_SLIPIF */
}

#if LWIP_DNS_APP && LWIP_DNS
static void
dns_found(const char *name, const ip_addr_t *addr, void *arg)
{
  LWIP_UNUSED_ARG(arg);
  LWIP_TRACE (1, "%s: %s\n", name, addr ? ipaddr_ntoa(addr) : "<not found>");
}

static void
dns_dorequest(void *arg)
{
  const char* dnsname = "3com.com";
  ip_addr_t dnsresp;
  LWIP_UNUSED_ARG(arg);

  if (dns_gethostbyname(dnsname, &dnsresp, dns_found, 0) == ERR_OK) {
    dns_found(dnsname, &dnsresp, 0);
  }
}
#endif

#if LWIP_LWIPERF_APP
static void
lwiperf_report (void *arg, enum lwiperf_report_type report_type,
                const ip_addr_t* local_addr, u16_t local_port,
                const ip_addr_t* remote_addr, u16_t remote_port,
                u32_t bytes_transferred, u32_t ms_duration,
                u32_t bandwidth_kbitpsec)
{
  LWIP_UNUSED_ARG(arg);
  LWIP_UNUSED_ARG(local_addr);
  LWIP_UNUSED_ARG(local_port);

  LWIP_TRACE (1, "IPERF report: type=%d, remote: %s:%d, total bytes: %lu, duration in ms: %lu, kbits/s: %lu\n",
               (int)report_type, ipaddr_ntoa(remote_addr), (int)remote_port, bytes_transferred, ms_duration, bandwidth_kbitpsec);
}
#endif

#if LWIP_MDNS_RESPONDER
static void
srv_txt(struct mdns_service *service, void *txt_userdata)
{
   err_t res = mdns_resp_add_service_txtitem(service, "path=/", strlen("path=/"));
   LWIP_ERROR("mdns add service txt failed\n", (res == ERR_OK), return);
   LWIP_UNUSED_ARG(txt_userdata);
}
#endif

/* This function initializes applications
 */
static void
apps_init(void)
{
#if LWIP_DNS_APP && LWIP_DNS
  /* wait until the netif is up (for dhcp, autoip or ppp) */
  sys_timeout(5000, dns_dorequest, NULL);
#endif

#if LWIP_CHARGEN_APP && LWIP_SOCKET
  chargen_init();
#endif

/*
 * ping out destination cmd option
 */
#if LWIP_PING_APP && LWIP_RAW && LWIP_ICMP
  if (lwip_cfg.app.ping.enable)
  {
    static ip_addr_t ping_addr;

    ipaddr_aton (lwip_cfg.app.ping.target_str, &ping_addr);
    ping_init (&ping_addr);
  }
#endif

#if LWIP_NETBIOS_APP && LWIP_UDP
  netbiosns_init();
  #ifndef NETBIOS_LWIP_NAME
    #if LWIP_NETIF_HOSTNAME
      netbiosns_set_name(netif_default->hostname);
    #else
      netbiosns_set_name("NETBIOSLWIPDEV");
    #endif
  #endif
#endif

#if LWIP_HTTPD_APP && LWIP_TCP
  #if LWIP_HTTPD_APP_NETCONN
    if (lwip_cfg.app.httpd_netconn.enable)
       http_server_netconn_init();
  #else
    #if defined(LWIP_HTTPD_EXAMPLE_CUSTOMFILES) && LWIP_HTTPD_EXAMPLE_CUSTOMFILES && defined(LWIP_HTTPD_EXAMPLE_CUSTOMFILES_ROOTDIR)
     fs_ex_init(LWIP_HTTPD_EXAMPLE_CUSTOMFILES_ROOTDIR);
    #endif
  #endif

  if (lwip_cfg.app.httpd.enable)
  {
    httpd_init();
#if defined(LWIP_HTTPD_EXAMPLE_SSI_SIMPLE) && LWIP_HTTPD_EXAMPLE_SSI_SIMPLE
    ssi_ex_init();
#endif
#if (HTTPD_ENABLE_HTTPS) && 0
    httpd_inits (&altcp_https);  /* \todo */
#endif
  }
#endif

#if LWIP_MDNS_RESPONDER
  mdns_resp_init();
  mdns_resp_add_netif(netif_default, "lwip");
  mdns_resp_add_service(netif_default, "lwipweb", "_http", DNSSD_PROTO_TCP, HTTPD_SERVER_PORT, srv_txt, NULL);
#endif

#if LWIP_NETIO_APP && LWIP_TCP
  netio_init();
#endif

#if LWIP_RTP_APP && LWIP_SOCKET && LWIP_IGMP
  rtp_init();
#endif

#if LWIP_SNTP_APP && LWIP_SOCKET
  if (lwip_cfg.app.sntp.enable && lwip_cfg.app.sntp.server_str)
  {
    sntp_setoperatingmode (lwip_cfg.app.sntp.mode);
    sntp_init();
  }
#endif

#if LWIP_SHELL_APP && LWIP_NETCONN
  if (lwip_cfg.app.shell.enable)
     shell_init();
#endif

#if LWIP_TCPECHO_APP
  #if LWIP_NETCONN && defined(LWIP_TCPECHO_APP_NETCONN)
    tcpecho_init();
  #else
    tcpecho_raw_init();
  #endif
#endif

#if LWIP_SNMP && LWIP_UDP
  snmp_trap_dst_ip_set (0,&trap_addr);
  snmp_trap_dst_enable (0,trap_flag);
  snmp_mib2_set_syscontact (syscontact_str, &syscontact_len, sizeof(syscontact_str));
  snmp_mib2_set_syslocation (syslocation_str, &syslocation_len, sizeof(syslocation_str));
//snmp_set_snmpenableauthentraps (&snmpauthentraps_set);

  #if LWIP_SNMP_V3
  snmpv3_dummy_init();
  #endif

  snmp_init();
#endif

#if LWIP_UDPECHO_APP && LWIP_NETCONN
  udpecho_init();
#endif

#if LWIP_LWIPERF_APP
  lwiperf_start_tcp_server_default(lwiperf_report, NULL);
#endif

#if LWIP_SOCKET_EXAMPLES_APP && LWIP_SOCKET
  socket_examples_init();
#endif

#if LWIP_UDP
  if (lwip_cfg.app.tftp.enable)
     tftp_init_client (&tftp);
#endif

#ifdef LWIP_APP_INIT
  LWIP_APP_INIT();
#endif
}

/* This function initializes this lwIP test. When NO_SYS=1, this is done in
 * the main_loop context (there is no other one), when NO_SYS=0, this is done
 * in the tcpip_thread context
 */
static void
test_init(void * arg)
{
#if NO_SYS
  LWIP_UNUSED_ARG(arg);
#else
  sys_sem_t *init_sem;
  LWIP_ASSERT("arg != NULL", arg != NULL);
  init_sem = (sys_sem_t*)arg;
#endif

  /* init randomizer again (seed per thread) */
  srand((unsigned int)time(0));

  /* init network interfaces */
  msvc_netif_init();

  /* init apps */
  apps_init();

#if !NO_SYS
  sys_sem_signal(init_sem);
#endif /* !NO_SYS */
}

static void
show_menu (void)
{
  sys_prot_t l = sys_arch_protect();

  puts ("m: print memory stats.");
  puts ("n: print TCP netstat.");
  puts ("p: print mem-pool stats.");
  puts ("S: print system stats.");
  puts ("s: print protocol stats.");
  puts ("h: this help.");
  puts ("i: print IPv4/IPv6 address, network-mask, gateway.");
  puts ("q: quit program.");
  sys_arch_unprotect (l);
}

static void
show_network_info (void)
{
  puts ("Not implemented yet.");

#if LWIP_IPV6
  puts ("IPv6 info:\n"
        "  linklocal address: ");
  ip6_addr_debug_print (0xFFFFFFFF, netif_ip6_addr(&netif, 0));
#else
  puts ("IPv6 not compiled in.");
#endif

}

static void
proto_stats_display (void)
{
  LINK_STATS_DISPLAY();
  ETHARP_STATS_DISPLAY();
  IPFRAG_STATS_DISPLAY();
  IP6_FRAG_STATS_DISPLAY();
  IP_STATS_DISPLAY();
  ND6_STATS_DISPLAY();
  IP6_STATS_DISPLAY();
  IGMP_STATS_DISPLAY();
  MLD6_STATS_DISPLAY();
  ICMP_STATS_DISPLAY();
  ICMP6_STATS_DISPLAY();
  UDP_STATS_DISPLAY();
  TCP_STATS_DISPLAY();
}

#if (LWIP_IPV4)
  #define ADDR_NTOA(x) ip4addr_ntoa ((const ip4_addr_t*)(x))
#elif (LWIP_IPV6)
  #define ADDR_NTOA(x) ip6addr_ntoa ((const ip6_addr_t*)(x))
#else
  #error "What?"
#endif

#define NETSTAT_FMT  "%s    %-20s  %-20s %s\n"

static void
show_tcp (const struct tcp_pcb *tcp)
{
  char local [60];
  char remote [60];
  char buf [10];

  strcpy (local, ADDR_NTOA(&tcp->local_ip));
  strcat (local, ":");
  strcat (local, itoa(tcp->local_port,buf,10));
  if (tcp->state <= LISTEN)
     strcpy (remote, "-");
  else
  {
    strcpy (remote, ADDR_NTOA(&tcp->remote_ip));
    strcat (remote, ":");
    strcat (remote, itoa(tcp->remote_port,buf,10));
  }
  printf (NETSTAT_FMT, "tcp", local, remote, tcp_debug_state_str(tcp->state));
}

static void
show_udp (const struct udp_pcb *udp)
{
  char local [60];
  char remote [60];
  char buf [10];

  strcpy (local, ADDR_NTOA(&udp->local_ip));
  strcat (local, ":");
  strcat (local, itoa(udp->local_port,buf,10));
  strcpy (remote, ADDR_NTOA(&udp->remote_ip));
  strcat (remote, ":");
  strcat (remote, itoa(udp->remote_port,buf,10));
  printf (NETSTAT_FMT, "udp", local, remote, "NONE");
}

static void
netstat_display (void)
{
  const struct tcp_pcb *tcp;
  const struct udp_pcb *udp;
  int   i;

  puts ("Proto  Local                 Remote               Status");

  for (i = 0; i < LWIP_ARRAYSIZE(tcp_pcb_lists); i++)
      for (tcp = *tcp_pcb_lists[i]; tcp; tcp = tcp->next)
          show_tcp (tcp);

  for (udp = udp_pcbs; udp; udp = udp->next)
      show_udp (udp);
}


#if defined(_MSC_VER) && defined(_DEBUG)
static _CrtMemState last_state;

static void crtdbug_exit (void)
{
  _CrtMemDumpAllObjectsSince (&last_state);
  _CrtMemDumpStatistics (&last_state);
#if 0
  _CrtCheckMemory();
  _CrtDumpMemoryLeaks();
#endif
}

static void
crtdbug_init (void)
{
  _CrtSetReportFile (_CRT_WARN, _CRTDBG_FILE_STDERR);
  _CrtSetReportMode (_CRT_WARN, _CRTDBG_MODE_FILE);
  _CrtSetDbgFlag (_CRTDBG_LEAK_CHECK_DF     |
                  _CRTDBG_DELAY_FREE_MEM_DF |
                  _CRTDBG_ALLOC_MEM_DF      |
                  _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG));
  _CrtMemCheckpoint (&last_state);
  atexit (crtdbug_exit);
}
#endif

/* This is somewhat different to other ports: we have a main loop here:
 * a dedicated task that waits for packets to arrive. This would normally be
 * done from interrupt context with embedded hardware, but we don't get an
 * interrupt in windows for that :-)
 */
static void
main_loop (void)
{
#if !NO_SYS
  err_t err;
  sys_sem_t init_sem;
#endif

#if USE_PPP
  int count;
#if !USE_ETHERNET
  u8_t rxbuf[1024];
#endif
  volatile int callClosePpp = 0;
#endif

/* Prevent an 'abort() GUI' popping up.
 */
#if defined(_MSC_VER) && defined(_DEBUG)
  _set_abort_behavior (0, _WRITE_ABORT_MSG);
  crtdbug_exit();
#endif

  /* initialize lwIP stack, network interfaces and applications */
#if NO_SYS
  lwip_init();
  test_init(NULL);
#else
  err = sys_sem_new(&init_sem, 0);
  LWIP_ASSERT("failed to create init_sem", err == ERR_OK);
  tcpip_init(test_init, &init_sem);

  /* we have to wait for initialization to finish before
   * calling update_adapter()! */
  sys_sem_wait(&init_sem);
  sys_sem_free(&init_sem);
#endif

#if (LWIP_SOCKET || LWIP_NETCONN) && LWIP_NETCONN_SEM_PER_THREAD
  netconn_thread_init();
#endif

  show_menu();
  fflush (stdout);

  /* MAIN LOOP for driver update (and timers if NO_SYS) */
  while (1)
  {
#if NO_SYS
    /* handle timers (already done in tcpip.c when NO_SYS=0) */
    sys_check_timeouts();
#endif

#if USE_ETHERNET
#if !PCAPIF_RX_USE_THREAD
    /* check for packets and link status*/
    pcapif_poll(&netif);
    /* When pcapif_poll comes back, there are not packets, so sleep to
       prevent 100% CPU load. Don't do this in an embedded system since it
       increases latency! */
    sys_msleep(1);
#else
    sys_msleep(50);
#endif

#else  /* 'USE_ETHERNET' not defined */
#if USE_PPP
    /* try to read characters from serial line and pass them to PPPoS */
    count = sio_read(ppp_sio, (u8_t*)rxbuf, 1024);
    if(count > 0)
    {
      pppos_input(ppp, rxbuf, count);
    }
    else
#endif
    {
      /* nothing received, give other tasks a chance to run */
      sys_msleep(1);
    }
#endif

#if USE_SLIPIF
    slipif_poll(&slipif1);
#if USE_SLIPIF > 1
    slipif_poll(&slipif2);
#endif
#endif

#if ENABLE_LOOPBACK && !LWIP_NETIF_LOOPBACK_MULTITHREADING
    /* check for loopback packets on all netifs */
    netif_poll_all();
#endif

#if USE_PPP
    {
      int do_hup = 0;
      if(do_hup)
      {
        ppp_close(ppp, 1);
        do_hup = 0;
      }
    }
    if(callClosePpp && ppp)
    {
      /* make sure to disconnect PPP before stopping the program... */
      callClosePpp = 0;
#if NO_SYS
      ppp_close(ppp, 0);
#else
      pppapi_close(ppp, 0);
#endif
      ppp = NULL;
    }
#endif /* USE_PPP */

    if (_kbhit() /* || lwip_win32_keypressed(NULL) */)
    {
      int ch;

    //MessageBeep (MB_OK);

      ch = _getch();

      if (ch == 'S')
      {
        puts ("-----------------------------------------");
        stats_display_sys (&lwip_stats.sys);
      }
      else if (ch == 's')
      {
        puts ("-----------------------------------------");
        proto_stats_display();
      }
      else if (ch == 'm')
      {
        puts ("-----------------------------------------");
        MEM_STATS_DISPLAY();
      }
      else if (ch == 'p')
      {
        int i;
        puts ("-----------------------------------------");
        for (i = 0; i < MEMP_MAX; i++)
            MEMP_STATS_DISPLAY (i);
      }
      else if (ch == 'n')
      {
        puts ("--------------------------------------------------------------");
        netstat_display();
      }
      else if (ch == 'i')
      {
        puts ("-----------------------------------------");
        show_network_info();
      }
      else if (ch == 'h')
      {
        puts ("-----------------------------------------");
        show_menu();
      }
      else if (ch == 'q')
        break;

      printf ("LwIP> ");
      fflush (stdout);
    }
  }

#if USE_PPP
  if (ppp)
  {
    u32_t started;
    LWIP_TRACE (1, "Closing PPP connection...\n");
    /* make sure to disconnect PPP before stopping the program... */
#if NO_SYS
    ppp_close(ppp, 0);
#else
    pppapi_close(ppp, 0);
#endif
    ppp = NULL;

    /* Wait for some time to let PPP finish... */
    started = sys_now();
    do
    {
#if USE_ETHERNET && !PCAPIF_RX_USE_THREAD
      pcapif_poll(&netif);
#else
      sys_msleep(50);
#endif
      /* @todo: need a better check here: only wait until PPP is down */
    }
    while (sys_now() - started < 5000);
  }
#endif /* USE_PPP */

#if (LWIP_SOCKET || LWIP_NETCONN) && LWIP_NETCONN_SEM_PER_THREAD
  netconn_thread_cleanup();
#endif

#if USE_ETHERNET
  /* release the pcap library... */
  pcapif_shutdown (&netif);
#endif
}

int main(int argc, char **argv)
{
  const char *e = getenv("LWIP_TRAP");

  lwip_cfg_init();

#if (LWIP_DEBUG)
  if (argc >= 2 && !strcmp(argv[1],"-v"))
  {
    print_USE_values();
    argc--;
    argv--;
  }
#endif

  if (e) {
#if (LWIP_SNMP && LWIP_IPV4)
    char ip_str[16];

    snmpauthentraps_set = 1;
    ipaddr_aton(e, &trap_addr);
    strncpy(ip_str, ip4addr_ntoa(&trap_addr),sizeof(ip_str));
    LWIP_TRACE (1, "SNMP trap destination %s\n", ip_str);
#endif
  }

#if USE_PPP && PPPOS_SUPPORT
  if(argc > 1) {
    sio_idx = (u8_t)atoi(argv[1]);
  }
  LWIP_TRACE (1, "Using serial port %d for PPP\n", sio_idx);
#endif

  /* no stdio-buffering, please! */
  setvbuf(stdout, NULL,_IONBF, 0);

  main_loop();

  return 0;
}
